#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#
# mechanics
#
#     Management, creation and destruction of parallel processes
#     on GUI applications using the python multiprocessing module.
#
# Written by E. A. Tacao <mailto@tacao.com.br>, (C) 2021, ..., 2024
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#     (1) Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#
#     (2) Redistributions in binary form must reproduce the above copyright
#     notice, this list of conditions and the following disclaimer in
#     the documentation and/or other materials provided with the
#     distribution.
#
#     (3)The name of the author may not be used to
#     endorse or promote products derived from this software without
#     specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT,
# INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
# STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
# IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#

__author__  = "E. A. Tacao <mailto |at| tacao.com.br>"
__date__    = "20 Apr 2024, 12:00 GMT-03:00"
__version__ = "0.9.2"

#-------------------------------------------------------------------------------------------------------------------------------

import logging, time

from datetime import datetime

from multiprocessing import Process, Lock, Manager
from six.moves import _thread

#-------------------------------------------------------------------------------------------------------------------------------

def handle_exception(exc_type, exc_value, exc_traceback):
    if issubclass(exc_type, KeyboardInterrupt): # Ignore KeyboardInterrupt
        sys.__excepthook__(exc_type, exc_value, exc_traceback)
        return
    logging.getLogger().error("Uncaught exception", exc_info=(exc_type, exc_value, exc_traceback))

#-------------------------------------------------------------------------------------------------------------------------------

class ProcX:
    """
    Internal use only. Wraps the processes so that they and Gear may exchange messages.
    """

    def __init__(self, ns, lock, keepGoing, tsid, obj, args, desc, dob, \
                 timeout=None, finishMethod="terminate"):
        self.ns = ns
        self.lock = lock
        self.keepGoing = keepGoing
        self.tsid = tsid
        self.obj = obj
        self.args = args
        self.desc = desc
        self.dob = dob
        self.timeout = timeout
        self.finishMethod = finishMethod
        self.isWaiting = True
        self.isStopping = False
        self.p = None


    def Comm(self, command, value=None):
        self.lock.acquire()
        cmd = (self.tsid, command, value)
        self.ns.append(cmd)
        self.lock.release()


    def ShouldKeepGoing(self):
        return self.keepGoing[self.tsid]


    def DontKeepGoing(self):
        self.keepGoing[self.tsid] = False


    def IsRunning(self):
        try:
            r = self.p.is_alive()
        except AttributeError:
            r = False
        return r


    def IsNotRunning(self):
        return not self.IsRunning()


    def IsWaiting(self):
        return self.isWaiting


    def IsNotWaiting(self):
        return not self.IsWaiting()


    def IsNotStopping(self):
        return not self.isStopping


    def IsTimedOut(self):
        if self.timeout is None:
            return False
        else:
            delta = time.time() - self.dob
            return delta >= self.timeout


    def Start(self):
        self.isWaiting = False
        args = (self.args) + (self.Comm, self.ShouldKeepGoing)
        self.p = Process(target=self.obj, args=args)
        self.p.name = self.tsid
        self.p.start()


    def Finish(self, finishMethod=""):
        self.DontKeepGoing(); self.isStopping = True
        finishMethod = finishMethod or self.finishMethod
        if self.p is not None:
            self.p.terminate() if finishMethod == "terminate" else self.p.kill()
            while True:
                if not self.p.is_alive():
                    break
                time.sleep(0.01)

#-------------------------------------------------------------------------------------------------------------------------------

class Gear:
    """
    Internal use only. Gear runs a separate thread, pools the processes for messages
    and send them to application using the supplied signals.
    """

    def __init__(self, parent, poll_interval):
        self.parent        = parent
        self.poll_interval = poll_interval


    def Start(self):
        self.keepGoing = self.running = True
        _thread.start_new_thread(self.Run, ())


    def Stop(self):
        self.keepGoing = False


    def IsRunning(self):
        return self.running


    def Run(self):
        while self.keepGoing:

            self.ProcessEvents()
            self.CheckRunning()
            self.CheckPool()

            time.sleep(self.poll_interval/100 if self.parent.ns else self.poll_interval)

        self.running = False


    def ProcessEvents(self):
        if self.parent.ns:
            self.parent.lock.acquire()
            first = self.parent.ns.pop(0)
            tsid, command, value = (list(first) + [None]*3)[:3]
            self.parent.lock.release()

            if command == "Finish":
                logging.info("process < %s > asked to finish; finishing it" % (tsid))
                self.parent.Stop(tsid)

            else:
                logging.info("message: " + repr((tsid, command, value)))
                self.parent.EmitSignal(tsid, command, value)


    def CheckRunning(self):
        processes = self.parent.processes
        for procx in processes:
            if procx.IsNotRunning() and procx.IsNotWaiting() and procx.IsNotStopping():
                logging.info("finishing process < %s >" % (procx.tsid))
                self.parent.Stop(procx.tsid)

            elif procx.IsTimedOut():
                logging.warning("process < %s > timed out; finishing it" % (procx.tsid))
                command = "Timeout"
                self.parent.EmitSignal(procx.tsid, command, None)
                self.parent.Stop(procx.tsid)


    def CheckPool(self):
        processes = self.parent.processes
        for procx in processes:
            if len(processes) < self.parent.max_proc:
                if procx.IsWaiting():
                    logging.info("creating process < %s >" % (procx.tsid))
                    procx.Start()
            else:
                break

#-------------------------------------------------------------------------------------------------------------------------------

class Mechanics:
    def __init__(self, max_proc=640, poll_interval=0.01, log_level="info"):

        """
        Mechanics extends a GUI application to help management, creation and destruction of parallel processes
        using the python multiprocessing module.

        Data passed between the application and processes must be pickable; all processes data
        are passed to the application via observer patterns (custom events (wxPython) or signals (PySide6)).

        The events receive the data: tsid, command and value:
            tsid (string):    the process string descriptor (tsid) given to the process on its creation
            command (string): a descriptive command to help the application manage the passed data. "Finish" may be passed
                              by the process to indicate it has finished its processing; "Timeout" is passed by the Gear
                              to indicate the process timed out and will stop.
            value:            data passed from the process to the application.
        <https://docs.python.org/3/library/multiprocessing.html>

        Constructor:
            Mechanics(max_proc, poll_interval, log_level)

            max_proc:       the maximum allowed number of concurrent processes
            poll_interval:  event loop polling interval (seconds)
            log_level:      logging verbosity (https://docs.python.org/3/library/logging.html#logging-levels)
        """

        logging.basicConfig(encoding="utf-8",
                                    format="%(asctime)s | %(levelname)s | %(module)s | %(funcName)s | %(message)s",
                                    datefmt="%d/%m/%Y %H:%M:%S", level=log_level.upper())

        self.max_proc       = max_proc
        self.processes      = []
        self.signals        = {}
        self.mgr            = Manager()
        self.ns             = self.mgr.list()
        self.lock           = self.mgr.Lock()
        self.keepGoing      = self.mgr.dict()
        self.finishMethod   = "terminate"

        self.gear = Gear(self, poll_interval)
        self.gear.Start()


    # Public methods -----------------------------------------------------------------------------------------------------------

    def SetMaxProc(self, max_proc):
        """
        SetMaxProc(max_proc)
            Changes the maximum allowed number of concurrent processes.
        """

        self.max_proc = max_proc


    def SetPollInterval(self, poll_interval):
        """
        SetPollInterval(poll_interval)
            Changes the polling interval when checking processes (in seconds).
        """

        self.gear.poll_interval = poll_interval


    def SetLogLevel(self, log_level):
        """
        SetLogLevel(log_level)
            Changes logging verbosity. log_level options are: "debug", "info", "warning", "error", "critical"
        """

        log = logging.getLogger()
        log.setLevel(log_level.upper())


    def SetDefaultFinishMethod(self, method="terminate"):
        """
        SetDefaultFinishMethod(method="terminate")
            Sets the default finishing method: can be either "terminate" or "kill".
            <https://docs.python.org/3/library/multiprocessing.html#multiprocessing.Process.terminate>
            <https://docs.python.org/3/library/multiprocessing.html#multiprocessing.Process.kill>
        """

        self.finishMethod = method


    def GetNumberOfProcesses(self):
        """
        GetNumberOfProcesses()
            Returns the length of self.processes, which is the number of running processes plus pooled processes.
        """

        return len(self.processes)


    def GetProcess(self, tsid):
        """
        GetProcess(tsid)
            Returns a ProcX object from a tsid, or None if the tsid is not on the current processes list.
        """

        for procx in self.processes:
            if procx.tsid == tsid:
                break
        else:
            procx = None

        return procx


    def EmitSignal(self, tsid, command, value):
        """
        EmitSignal(tsid, command, value)
            Emits the signal supplied with this tsid with tsid itself, command and value as args.
        """

        if tsid in self.signals.keys():
            self.signals[tsid].emit((tsid, command, value))
        else:
            logging.warning("error: process < %s > doesn't exist" % (tsid))


    def Start(self, obj, desc, timeout=None, signal=None, args=()):
        """
        Start(obj, desc, timeout=None, signal=None, args=())
            Starts a new callable process (obj).

            desc:       (string) an arbitrary small description of the process.
            timeout:    period of time that will be allowed to elapse before this process is stopped.
                        if None, there will be no timeout.
            signal:     the custom signal that should be bound to the application to fetch the process calls.
            args:       arguments to the process

            Returns, if the creation was successful, a string descriptor (tsid) containing the description
            and the creation timestamp. Returns None if the process already exists.
        """

        dob = time.time()
        timestamp = datetime.fromtimestamp(dob).strftime("%d/%m/%Y %H:%M:%S.%f")
        tsid = "%s | %s" % (desc, timestamp)

        if tsid in self.processes:
            logging.warning("error: process < %s > already exists" % (tsid))
            return None

        else:
            logging.info("pooling creation of process < %s >" % (tsid))

            self.keepGoing[tsid] = True
            pp = ProcX(self.ns, self.lock, self.keepGoing, tsid, obj, args, desc, dob, timeout)
            self.processes.append(pp)
            self.signals[tsid] = signal
            return tsid


    def AskToStop(self, tsid):
        """
        AskToStop(tsid):
            Asks nicely to stop a process by its string descriptor (tsid). If implemented on the
        process itself (by watching ShouldKeepGoing), it may end gracefully by calling Comm("Finish").
        """

        procx = self.GetProcess(tsid)
        if not procx:
            logging.warning("process < %s > doesn't exist" % (tsid))
        else:
            procx.DontKeepGoing() #, please.


    def Stop(self, tsid, method=""):
        """
        Stop(tsid, method="")
            Given a process string descriptor (tsid), stops (terminates or kills) the process,
            according to the method passed. If no stopping method is passed, it will use the default finish method
            currently set.
        """

        method = method or self.finishMethod

        if method == "terminate":
            mstr = "terminating"
        elif method == "kill":
            mstr = "killing"
        else:
            logging.warning("invalid stop method; should be 'terminate' or 'kill'; was: '%s'" % method)
            return

        procx = self.GetProcess(tsid)
        if not procx:
            logging.warning("process < %s > doesn't exist" % (tsid))
        else:
            logging.info("%s process < %s >" % (mstr, tsid))
            self.EmitSignal(tsid, "Stop", None)
            procx.Finish(method)
            try:
                if procx in self.processes:
                    self.processes.remove(procx)
                if procx.tsid in self.keepGoing.keys():
                    del self.keepGoing[procx.tsid]
            except (KeyError, ValueError):
                pass
            del(procx)


    def StopAll(self, method=""):
        """
        StopAll(method="")
            Stops (terminates or kills) all running processes according to the method passed.
            If no stopping method is passed, it will use the default finish method
            currently set.
        """

        method = method or self.finishMethod
        m = [[procx.tsid, method] for procx in self.processes]  # ensure correct gc
        [self.Stop(x, y) for x, y in m]


    def TurnOff(self, method=""):
        """
        TurnOff(method="")
            This method will stop all running processes and the gear thread. It should be called
            before exiting the QApplication (e.g, on closeEvent). If no stopping method is passed, it will use the
            default finish method currently set.
        """

        method = method or self.finishMethod

        logging.info("Sending messages to stop all processes (%s)..." % len(self.processes))
        self.StopAll(method)
        logging.info("...messages sent.")

        logging.info("Waiting for processes to stop...")
        while len(self.processes):
            time.sleep(0.01)
        logging.info("...processes stopped.")

        logging.info("Stopping gear...")
        self.gear.Stop()
        while self.gear.IsRunning():
            time.sleep(0.01)
        logging.info("...gear stopped.")


#
##
### eof
