#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#
# test program for mechanics
#
#     Management, creation and destruction of parallel processes
#     on GUI applications using the python multiprocessing module.
#
# Written by E. A. Tacao <mailto@tacao.com.br>, (C) 2021, ..., 2024
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#     (1) Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#
#     (2) Redistributions in binary form must reproduce the above copyright
#     notice, this list of conditions and the following disclaimer in
#     the documentation and/or other materials provided with the
#     distribution.
#
#     (3)The name of the author may not be used to
#     endorse or promote products derived from this software without
#     specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT,
# INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
# STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
# IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#

#-------------------------------------------------------------------------------------------------------------------------------

import sys, time, logging, multiprocessing, decimal

from mechanics import Mechanics

# warning! potential triggers ahead :P
from PySide6.QtCore import *
from PySide6.QtWidgets import *
from PySide6.QtGui import *
# also, 0 O = 0: GOTO 10            :P

#-------------------------------------------------------------------------------------------------------------------------------

def calculate_pi(num_decimals, *args):  # the Prof. Simon Plouffe way

    # We have to be able to signal data to the program that created a process with this funcion, and we also need
    # to be able to monitor if the program asked us nicely to stop the process (so that we can, actually, nicely stop).
    #
    # It may also be useful to run this code externally to the program, even from the command line perhaps, for
    # testing purposes.
    #
    # The following line provides all of that:

    Comm, ShouldKeepGoing = (print, lambda: True) if not args else args[-2:]

    decimal.getcontext().prec = num_decimals + 2
    pi = decimal.Decimal(0)
    for k in range(num_decimals):
        if ShouldKeepGoing():   # This is how we periodically monitor for mechanics.AskToStop(). This is not stricly required,
                                # but allows a process to perform some proper cleanup before stopping.

            pi += (decimal.Decimal(1)/(16**k))*((decimal.Decimal(4)/(8*k+1)) - (decimal.Decimal(2)/(8*k+4)) - \
                    (decimal.Decimal(1)/(8*k+5)) - (decimal.Decimal(1)/(8*k+6)))
            if k/1000 == int(k/1000):   # A low-pass filter is in order so that
                                        # we don't choke PySide's neither mechanic's mainloops.

                # Comm will make mechanics emit a signal containing a tuple (descriptor, command, value).
                # In this case, our descriptor, "Progress" and k.
                Comm("Progress", k)
        else:
            Comm("Interrupted") # Here the tuple will be our descriptor, "Interrupted" and None.
            break
    else:
        Comm("Result", pi) # Here the tuple will be our descriptor, "Result" and pi.

    # If the process is to finish, always finish it like this so that mechanics can properly be notified.
    #
    # Note that 'Finish' is a reserved word and should not be used elsewhere on Comm(). ('Timeout' and 'Stop'
    # are also reserved words and should not be used anywhere on Comm()).

    Comm("Finish")

#-------------------------------------------------------------------------------------------------------------------------------

# A QPlainTextEdit that receives logging messages.

class QTextEditLogger(logging.Handler):
    class Emitter(QObject):
        log = Signal(str)


    def __init__(self, parent):
        super().__init__()
        self.widget = QPlainTextEdit(parent)
        self.widget.setReadOnly(True)
        self.emitter = QTextEditLogger.Emitter()
        self.emitter.log.connect(self.widget.appendPlainText)


    def emit(self, record):
        msg = self.format(record)
        self.emitter.log.emit(msg)

#-------------------------------------------------------------------------------------------------------------------------------

class mainw(QMainWindow):

    # The object that will receive messages from our processes.
    mech_signal = Signal(object)

    def __init__(self):
        super().__init__()

        # Here we 'initialize' the mechanics module ---
        self.mech = Mechanics()
        self.mech_signal.connect(self.OnMech)
        # ---------------------------------------------

        self.setWindowTitle("My App")
        self.resize(QSize(400, 300))

        widget = QWidget()
        layout = QVBoxLayout()
        widget.setLayout(layout)
        self.setCentralWidget(widget)

        layout1 = QHBoxLayout()
        text = QLabel("Timeout: ")
        layout1.addWidget(text)
        self.to_text = QLineEdit("30")
        layout1.addWidget(self.to_text)
        layout1.addStretch(1)
        text = QLabel("Current processes: ")
        layout1.addWidget(text)
        self.no_proc = QLabel("0")
        layout1.addWidget(self.no_proc)
        layout.addLayout(layout1)

        layout2 = QHBoxLayout()
        text = QLabel("Total progress: ")
        layout2.addWidget(text)
        self.prog1 = QProgressBar()
        layout2.addWidget(self.prog1)
        layout.addLayout(layout2)

        line = QFrame(self)
        line.setFrameShape(QFrame.HLine)
        line.setFrameShadow(QFrame.Sunken)
        layout.addWidget(line)

        log_text = QTextEditLogger(self)
        logging.getLogger().addHandler(log_text)
        layout.addWidget(log_text.widget)

        line = QFrame(self)
        line.setFrameShape(QFrame.HLine)
        line.setFrameShadow(QFrame.Sunken)
        layout.addWidget(line)

        layout3 = QHBoxLayout()
        labels = ["Start", "Stop", "Stop, please!"]
        for label in labels:
            button = QPushButton(label)
            button.clicked.connect(lambda clicked=True,
                                   this_label=label: self.OnButtonClicked(this_label))
            layout3.addWidget(button)

        layout.addLayout(layout3)

        self.current_procs = {}


    def closeEvent(self, event):
        # Here we 'deinitialize' mechanics, safely killing all running processes (if any). It doesn't really need to be inside
        # a window's closeEvent method, but we must call TurnOff() when we are about to exit the GUI program nevertheless.

        self.mech.TurnOff()


    def OnButtonClicked(self, label):
        if label == "Start":
            timeout = int(self.to_text.text())

            # Launch a new process that will run the calculate_pi function; let its timeout be the number, in seconds,
            # typed in the 'Timeout' QLineEdit. Whenever the function communicates something, we will receive via
            # self.mech_signal (which was connected to our OnMech method on __init__). We also are passing a
            # tuple of arguments to the function.
            #
            # This returns a string descriptor (tsid) containing the description and the creation timestamp.

            tsid = self.mech.Start(calculate_pi, "pi Calc", timeout=timeout, signal=self.mech_signal, args=(5000, ))

            # In this example, each time we press the 'Start' button, we launch another process containing the
            # calculate_pi function. Here we add the descriptor to a dictionary, so we can keep track of the processes created.
            self.current_procs[tsid] = 0
            self.no_proc.setText(str(len(self.current_procs.keys())))

        elif label == "Stop":
            # In this example, each time we press the 'Stop' button, we stop the oldest process created that is still
            # running.
            if self.current_procs:
                first = list(self.current_procs.keys())[0]
                self.mech.Stop(first)

        elif label == "Stop, please!":
            # In this example, each time we press the 'Stop, please' button, we ask the oldest process created that is still
            # running to stop.
            if self.current_procs:
                first = list(self.current_procs.keys())[0]
                self.mech.AskToStop(first)


    def OnMech(self, t_c_v):
        # In this example, this is the slot that receives the signals produced by self.mech_signal. It can be named whatever
        # we want.

        # Note that it is also possible to have differente processes signaling messages to different slots: doing
        # that is trivial: the 'signal' keyword on mechanics.Start will have to receive the desired slot, and the desired
        # slot will obviously have to be created somewhere.

        tsid, command, value = t_c_v    # mechanics always signals a 3 item tuple: the descriptor, a 'command' and its 'value'.

        if tsid.startswith("pi Calc"):
            if command == "Result":
                logging.info("GUI: %s : ...5000 digits..." % command)
                self.UpdateProgress(tsid, 5000)
            elif command == "Progress":
                self.UpdateProgress(tsid, value)
            elif command == "Stop":
                del self.current_procs[tsid]
                self.no_proc.setText(str(len(self.current_procs.keys())))
            else:
                logging.info("GUI: %s : %s" % (command, value))


    def UpdateProgress(self, tsid, value):
        # Our QProgressBar shows the 'global' update progress measuring the progress of all running processes.
        self.current_procs[tsid] = value
        max_ = 5000 * len(self.current_procs); total_ = 0
        for v_ in self.current_procs.values():
            total_ += v_
        self.prog1.setRange(0, max_)
        self.prog1.setValue(total_)

#-------------------------------------------------------------------------------------------------------------------------------

if __name__ == "__main__":
    multiprocessing.set_start_method('spawn')   # level field between Linux and Windows
    app = QApplication([])
    window = mainw()
    window.show()
    app.exec()


#
##
### eof
